import asyncio
import itertools
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Tuple, Set, Optional

from croniter import croniter

from acron.job import Job, ScheduledJob, _job_context, JobContext

__all__ = ["Scheduler", "ScheduledJob"]


log = logging.getLogger("acron")


def enable_acron_debug_logs():
    log.addHandler(logging.StreamHandler())
    log.setLevel(logging.DEBUG)


if os.getenv("ACRON_DEBUG", "") == "TRUE":
    enable_acron_debug_logs()


ScheduledJobHandle = Tuple[ScheduledJob, asyncio.TimerHandle]


def croniter_sort_jobs(
    jobs: Set[Job], tz: timezone, n: int, offset: Optional[float]
) -> List[Tuple[float, Job]]:
    """
    Compute the n first jobs needed to be scheduled.
    If offset is defined then the jobs computed will be those that need to run after that time.
    """
    xs: List[Tuple[float, Job]] = []
    if offset:
        date = datetime.fromtimestamp(offset).astimezone(tz=tz)
    else:
        date = datetime.now(tz=tz)
    for job in jobs:
        xs.extend((d, job) for d in itertools.islice(croniter(job.schedule, date), n))

    new_jobs = sorted(xs, key=lambda i: i[0])
    last_job_time = new_jobs[n - 1][0]
    return list(itertools.takewhile(lambda i: i[0] <= last_job_time, new_jobs))


def cancel_old_jobs(
    generation: int, tasks: Dict[int, List[ScheduledJobHandle]]
) -> None:
    """
    Cancel all the scheduled jobs.
    """
    for scheduled_job, hs in tasks.get(generation, []):
        scheduled_job.event.set()
        hs.cancel()


def remove_completed_jobs(gen: int, tasks: Dict[int, List[ScheduledJobHandle]]) -> None:
    """
    Removes the scheduled jobs already completed.
    """
    xs = []
    for scheduled_job, hs in tasks.get(gen, []):
        if scheduled_job.event.is_set():
            log.debug(
                "[scheduler id=%s] Removing completed job %s",
                scheduled_job.id,
                scheduled_job.job.name,
            )
            continue
        xs.append((scheduled_job, hs))
    tasks[gen] = xs


def schedule_jobs(
    tasks: Dict[int, List[ScheduledJobHandle]],
    jobs: Set[Job],
    offset: Optional[float],
    n: int,
    generation: int,
    tz: timezone,
    dry_run: bool,
) -> float:
    """
    Schedule the next jobs from the defined jobs.
    It returns the timestamp for the last job scheduled.
    """
    loop = asyncio.get_running_loop()
    new_jobs = croniter_sort_jobs(jobs, tz, n, offset)
    for when, job in new_jobs:
        delta = datetime.fromtimestamp(when).astimezone(tz=tz) - datetime.now(tz=tz)
        # We need to create function here to capture the lexical context of
        # the parameters
        scheduled_job = ScheduledJob(job=job, tz=tz, when=when, dry_run=dry_run)
        # We need to call the lambda here because the coroutine needs to be
        # created when the job is launched, otherwise no one is awaiting it
        # and python complains.
        h = loop.call_later(
            delta / timedelta(seconds=1),
            lambda sj: asyncio.create_task(sj.run()),
            scheduled_job,
        )
        tasks[generation].append((scheduled_job, h))
    return new_jobs[-1][0]


def show_scheduled_jobs_info(
    scheduled_jobs: Dict[int, List[ScheduledJobHandle]], gen: int
) -> None:
    """
    Show information for the next jobs scheduled.
    """
    if not scheduled_jobs.get(gen, []):
        log.info("[scheduler] No jobs scheduled yet (generation %d)", gen)
        return
    log.info("[scheduler] Next jobs scheduled (generation %d):", gen)
    for scheduled_job, _ in scheduled_jobs.get(gen, []):
        if not scheduled_job.event.is_set():
            job_str: Optional[str] = None
            if scheduled_job.job.show:
                token = _job_context.set(JobContext(scheduled_job))
                try:
                    job_str = scheduled_job.job.show(scheduled_job.job.data)
                finally:
                    _job_context.reset(token)

            log.info(
                "[scheduler]  * [%s] %s at %s - %s",
                scheduled_job.id,
                scheduled_job.job.name or "unnamed job",
                scheduled_job.cron_date,
                job_str or "no job data",
            )


class Scheduler:
    def __init__(
        self,
        tz: Optional[timezone] = None,
        *,
        now: Optional[datetime] = None,
        dry_run: bool = False,
        scheduled_jobs_size: int = 32,
    ) -> None:
        self._jobs_queue: "Optional[asyncio.Queue[Set[Job]]]" = None
        self._tz = tz if tz is not None else timezone.utc
        self._stop_event: Optional[asyncio.Event] = None
        self._scheduler_future: Optional["asyncio.Future[None]"] = None
        self._defined_jobs: Optional[Set[Job]] = None
        self._scheduled_jobs: Dict[int, List[ScheduledJobHandle]] = {}
        self._scheduled_jobs_size = scheduled_jobs_size
        self._generation = 0
        self._last_job_time: Optional[float] = None
        self._last_scheduled_info = now or datetime.now()
        self._last_scheduled_delay = 3600
        self._dry_run = dry_run

    def process_jobs_update(
        self, new_jobs: Set[Job], *, now: Optional[datetime] = None
    ) -> None:
        if now is None:
            now = datetime.now()
        new_jobs_enabled = {job for job in new_jobs if job.enabled}
        if (
            self._defined_jobs and self._defined_jobs != new_jobs_enabled
        ) or not self._defined_jobs:
            cancel_old_jobs(self._generation, self._scheduled_jobs)
            self._defined_jobs = new_jobs_enabled
            if self._generation > 0:
                del self._scheduled_jobs[self._generation]
            self._generation = self._generation + 1
            self._scheduled_jobs[self._generation] = []
            self._last_job_time = None
        elif (now - self._last_scheduled_info).seconds > self._last_scheduled_delay:
            show_scheduled_jobs_info(
                scheduled_jobs=self._scheduled_jobs, gen=self._generation
            )
            self._last_scheduled_info = now

    def schedule_jobs(self) -> None:
        total_jobs = sum(
            1 for scheduled_job, _ in self._scheduled_jobs[self._generation]
        )
        num_active_jobs = sum(
            1
            for scheduled_job, _ in self._scheduled_jobs[self._generation]
            if not scheduled_job.event.is_set()
        )
        log.debug(
            "[scheduler] Number of active jobs: %d/%d", num_active_jobs, total_jobs
        )
        free_slots = self._scheduled_jobs_size - num_active_jobs
        log.debug("[scheduler] Number of free slots: %d", free_slots)
        if free_slots > 0 and self._defined_jobs:
            # we did not update the generation, keep queueing jobs,
            self._last_job_time = schedule_jobs(
                tasks=self._scheduled_jobs,
                jobs=self._defined_jobs,
                n=free_slots,
                generation=self._generation,
                tz=self._tz,
                offset=self._last_job_time,
                dry_run=self._dry_run,
            )
            show_scheduled_jobs_info(
                scheduled_jobs=self._scheduled_jobs, gen=self._generation
            )

    def scheduled_jobs(self) -> List[ScheduledJob]:
        return [
            scheduled_job for scheduled_job, _ in self._scheduled_jobs[self._generation]
        ]

    async def _run(self) -> None:
        assert self._stop_event is not None
        assert self._jobs_queue is not None
        while not self._stop_event.is_set():
            remove_completed_jobs(self._generation, self._scheduled_jobs)
            # Wait for new jobs, make sure we don't wait the whole timeout if stopped
            get_future = asyncio.ensure_future(self._jobs_queue.get())
            stop_future = asyncio.ensure_future(self._stop_event.wait())
            done, pending = await asyncio.wait(
                (get_future, stop_future),
                timeout=10.0,
                return_when=asyncio.FIRST_COMPLETED,
            )
            for future in pending:
                future.cancel()
            for future in done:
                if future == get_future:
                    jobs = get_future.result()
                    self.process_jobs_update(jobs)
            self.schedule_jobs()

    def start(self) -> None:
        assert self._jobs_queue is None
        assert self._stop_event is None
        self._jobs_queue = asyncio.Queue()
        self._stop_event = asyncio.Event()
        self._scheduler_future = asyncio.ensure_future(self._run())

    def stop(self) -> None:
        assert self._stop_event is not None
        self._stop_event.set()

    async def wait(self) -> None:
        assert self._scheduler_future is not None
        await self._scheduler_future
        self._stop_event = None
        self._scheduler_future = None
        self._jobs_queue = None

    def cleanup(self) -> None:
        for jobs in self._scheduled_jobs.values():
            for job, handle in jobs:
                handle.cancel()

    async def update_jobs(self, jobs: Set[Job]) -> None:
        assert self._jobs_queue is not None
        await self._jobs_queue.put(jobs)

    async def __aenter__(self) -> "Scheduler":
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.stop()
        await self.wait()
        self.cleanup()

    @property
    def running(self) -> bool:
        return self._scheduler_future is not None and not self._scheduler_future.done()



import logging
from datetime import timezone
from typing import Iterable, Optional

from acron.job import Job, SimpleJob


__all__ = ["__version__", "run", "Job", "SimpleJob", "logger"]


__version__ = "0.2.1"


logger = logging.getLogger("acron")


async def run(jobs: Iterable[Job], tz: Optional[timezone] = None) -> None:
    """
    Run given jobs with given timezone forever.
    """
    from acron.scheduler import Scheduler

    async with Scheduler(tz=tz) as _scheduler:
        await _scheduler.update_jobs(set(jobs))
        await _scheduler.wait()



import asyncio
import contextvars
import dataclasses
import logging
import uuid
from dataclasses import dataclass, field
from datetime import timezone, datetime
import time
from typing import Optional, Callable, Awaitable, Generic, TypeVar, Tuple

__all__ = ["Job", "SimpleJob", "ScheduledJob", "JobContext", "job_context"]


log = logging.getLogger("acron")
TaskT = TypeVar("TaskT")


@dataclass(frozen=True)
class Job(Generic[TaskT]):
    schedule: str
    data: TaskT
    func: Callable[[TaskT], Awaitable[None]] = field(compare=False)
    show: Optional[Callable[[TaskT], str]] = field(default=None, compare=False)
    name: Optional[str] = None
    enabled: bool = True

    @staticmethod
    def simple(
        schedule: str,
        func: Callable[[], Awaitable[None]],
        *,
        show: Optional[Callable[[], str]] = None,
        name: Optional[str] = None,
        enabled: bool = True
    ) -> "Job[Tuple[()]]":
        return Job[Tuple[()]](
            schedule=schedule,
            func=lambda _: func(),
            data=(),
            name=name,
            show=lambda _: show() if show else "",
            enabled=enabled,
        )


SimpleJob = Job.simple


def cron_date(timestamp: float, tz: timezone) -> str:
    fmt = "%Y-%m-%d %H:%M:%S %Z%z"
    return datetime.fromtimestamp(timestamp).astimezone(tz=tz).strftime(fmt)


@dataclasses.dataclass(frozen=True)
class ScheduledJob(Generic[TaskT]):
    job: Job[TaskT]
    tz: timezone
    when: float
    dry_run: bool
    event: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

    async def run(self) -> None:
        log.debug("[scheduler id=%s] Running scheduled job %s", self.id, self.job.name)
        start = time.monotonic()
        token = _job_context.set(JobContext(self))
        try:
            # mypy gets confused because we are calling a function but
            # it looks like we are calling a method.
            await self.job.func(self.job.data)  # type: ignore
        finally:
            _job_context.reset(token)
            self.event.set()
            log.debug(
                "[scheduler id=%s] Done running job %s after %.1f seconds",
                self.id,
                self.job.name,
                time.monotonic() - start,
            )

    @property
    def cron_date(self) -> str:
        return cron_date(timestamp=self.when, tz=self.tz)


@dataclasses.dataclass(frozen=True)
class JobContext:
    scheduled_job: ScheduledJob


_job_context: contextvars.ContextVar[JobContext] = contextvars.ContextVar(
    "acron_job_context"
)


def job_context() -> JobContext:
    return _job_context.get()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acron']

package_data = \
{'': ['*']}

install_requires = \
['croniter>=1.0.15,<2.0.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

setup_kwargs = {
    'name': 'acron',
    'version': '0.2.1',
    'description': 'Lightweight scheduler',
    'long_description': 'Lightweight scheduler for python asyncio\n\nBased on croniter to support the crontab syntax.\n\n============\nInstallation\n============\n\nInstalling acron.\n\n.. code:: shell\n\n    $ pip install acron\n\n=====\nUsage\n=====\n\nTo get started you need at least one job.\nUse the top level ``acron.run`` function for simple scheduling.\n\n\n.. code:: python\n\n    import asyncio\n    import acron\n\n    async def do_the_thing():\n        print("Doing the thing")\n\n    do_thing = acron.SimpleJob(\n        name="Do the thing",\n        schedule="0/1 * * * *",\n        func=do_the_thing,\n    )\n\n    asyncio.run(acron.run({do_thing}))\n\n\nFor more advanced use cases, the ``Scheduler`` class can be used as async context manager.\nCall ``scheduler.wait()`` to keep it running forever.\nTo submit jobs call ``scheduler.update_jobs(jobs)`` with the complete set of jobs.\n\nRunning a simple example running a function every hour...\n\n\n.. code:: python\n\n    import asyncio\n    import dataclasses\n\n    from acron.scheduler import Scheduler, Job\n\n    @dataclasses.dataclass(frozen=True)\n    class ThingData:\n        foo: bool\n\n    async def do_the_thing(data: ThingData):\n        print(f"Doing the thing {data}")\n\n    async def run_jobs_forever():\n        do_thing = Job[ThingData](\n            name="Do the thing",\n            schedule="0/1 * * * *",\n            data=ThingData(True),\n            func=do_the_thing,\n        )\n\n        async with Scheduler() as scheduler:\n            await scheduler.update_jobs({do_thing})\n            await scheduler.wait()\n\n    if __name__ == "__main__":\n        try:\n            asyncio.run(run_jobs_forever())\n        except KeyboardInterrupt:\n            print("Bye.")\n\n\n\n\nSpecifying a timezone\n----------------------\n\nFor python 3.9+ you can use the standard library\'s ``zoneinfo`` module to specify a timezone.\n\n.. code:: python\n\n    import zoneinfo\n\n    async with Scheduler(tz=zoneinfo.ZoneInfo("Europe/Berlin")) as scheduler:\n        ...\n\n\n\nFor earlier python versions you can use a third party library like ``pytz``.\n\n.. code:: python\n\n    import pytz\n\n    async with Scheduler(tz=pytz.timezone("Europe/Berlin")) as scheduler:\n        ...\n\n\nJob context\n-----------\n\nIt is possible to retrieve the context for the scheduled job from the running\njob function using ``job_context()``. This returns a ``JobContext`` containing\na reference to the ``ScheduledJob``. The ``job_context()`` function is implemented\nusing contextvars to provide the correct context to the matching asyncio task.\n\n.. code:: python\n\n    async def my_job_func():\n        job_id = acron.job_context().scheduled_job.id\n        job_name = acron.job_context().scheduled_job.job.name\n        print(f"Running job {job_id!r}, scheduled with id {job_id}")\n\n\n=================\nLocal development\n=================\n\nThe project uses poetry to run the test, the linter and to build the artifacts.\n\nThe easiest way to start working on acron is to use docker with the dockerfile\nincluded in the repository (manual usage of poetry is explained here:\nhttps://python-poetry.org/docs/ ).\n\nTo use docker, first generate the docker image. Run this command from the top\nlevel directory in the repository:\n\n.. code-block:: console\n\n   docker build -t acron-builder -f docker/Dockerfile .\n\nNow you can use it to build or run the linter/tests:\n\n.. code-block:: console\n\n    $ alias acron-builder="docker run --rm -it -v $PWD/dist:/build/dist acron-builder"\n\n    $ acron-builder run pytest tests\n    =============================================================================================== test session starts ================================================================================================\n    platform linux -- Python 3.9.7, pytest-5.4.3, py-1.10.0, pluggy-0.13.1\n    rootdir: /build\n    plugins: asyncio-0.15.1\n    collected 4 items\n    tests/test_acron.py ....                                                                                                                                                                                     [100%]\n    ================================================================================================ 4 passed in 0.04s =================================================================================================\n\n    $ acron-builder build\n    Building acron (0.1.0)\n      - Building sdist\n      - Built acron-0.1.0.tar.gz\n      - Building wheel\n      - Built acron-0.1.0-py3-none-any.whl\n\n    $ ls dist\n    acron-0.1.0-py3-none-any.whl  acron-0.1.0.tar.gz\n\n\n=========\nDebugging\n=========\n\nDebug logging can be enabled by setting the ``ACRON_DEBUG`` environment variable to ``TRUE``.\n\n',
    'author': 'Aitor Iturri',
    'author_email': 'aitor.iturri@appgate.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/appgate/acron',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
